; general discrimination tree climbing and building

; -----------------------------
; data types and definitions

; a node has three properties ASSOCIATED-ATOM, SEXP and ALIST
(defun node-sexp (node)
   (get node 'sexp))
; set sexp uses ADDPROP, clobber-sexp uses PUTPROP (resets)
(defun node-set-sexp (node sexp)
    (addprop node sexp 'sexp)
    (dirty-node node))
(defun node-clobber-sexp (node sexp)
    (putprop node sexp 'sexp)
    (dirty-node node))
(defun node-reset-sexp (node)
    (putprop node nil 'sexp)
    (dirty-node node))

(defun node-alist (node)
  (get node 'alist))
(defun node-set-alist (node alist)
  (addprop node alist 'alist)
  (dirty-node node))
(defun node-reset-alist (node)
  (putprop node nil 'alist)
  (dirty-node node))

(defun node-associated-atom (node)
  (get node 'associated-atom))
(defun node-set-associated-atom (node associated-atom)
  (addprop node associated-atom 'associated-atom)
  (dirty-node node))
(defun node-reset-associated-atom (node)
  (putprop node nil 'associated-atom)
  (dirty-node node))

; ---------------------------
;  necessary utilities

; add to last is essentially a destructive append
(defun add-to-last (lst itm)
    (rplacd (last lst) (ncons itm)))

(defun addprop (atm val prop)
   (putprop atm (cons val (get atm prop)) prop))

(defun make-node (sexp alist associated-atom)
 (let ((node-name (newsym 'NODE)))
      (putprop node-name sexp 'sexp)
      (putprop node-name alist 'alist)
      (putprop node-name associated-atom 'associated-atom)
      (dirty-node node)
      node-name))

; -------------------
; Index adds an arbitrary s-expression to a tree
(defun index (sexp node associated-atom)
    (let ((leaf-node (establish-links sexp node)))
	 (node-clobber-sexp leaf-node sexp)
	 (cond ((node-associated-atom leaf-node))   
                	        ; if associated-atom is there return it
	       (t (node-set-associated-atom leaf-node associated-atom)))))
 				; otherwise create one with the called value

; ------------- NOTE THE DOUBLE RECURSION ---------
; -------------                           ---------
(defun establish-links (item node)
    (cond ((null item) (establish-link '*NIL* node))
	  ((atom item)(establish-link item node))
	  ((establish-links 
	       (cdr item)
	       (establish-links 
		   (car item)
		   (establish-link '*CONS* node))))))

(defun establish-link (sexp node)
   (let ((auxil (assoc sexp (node-alist node))))
    (cond (auxil (cadr auxil))
	  (t (let ((new-node (make-node sexp nil nil)))
		  (add-to-alist node new-node sexp)
		  new-node)))))


(defun add-to-alist (node new-node sexp)
     (node-set-alist node (list sexp new-node)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun DISPLAY (node)
  (display-tree node 0))

(defun display-tree (node level)
  (msg (B level) "Node: " node
       " alist: " (node-alist node) N
       (B level) "sexp: " (node-sexp node)
       " associated-atom: " (node-associated-atom node) N)
  (cond
    ((null (node-alist node)) nil)
    (t (mapcar #'(lambda (link)
		   (display-tree (link-node link) (add1 level)))
	       (node-alist node))
       nil)))

(defun sweep ()
  (clean-tree '*KB*))

(defun clean-tree (node)
  (cond ((get node 'dirty)
	 (send-node node)
	 (putprop node nil 'dirty)))
  (cond
    ((null (node-alist node)) nil)
    (t (mapcar #'(lambda (link)
		   (clean-tree (link-node link)))
	       (node-alist node))
       nil)))

(defun dirty-node (node)
  (putprop node t 'dirty))

(defun send-node (node)
  (ps-make-node node (node-sexp node)
    (node-associated-atom node)
    (node-alist node)))

(setq StartPS "")
(setq EndPS "")

(defun ps-make-node (name sexp associated-atom alist)
  (msg "% Sending " name "...")
  (msg StartPS "userdict /" name " [" N)
  (msg "  " (list 'Name= name) " {" (ncons name) " print}" N
       "  " (list 'Sexp= sexp) " {" (ncons sexp) " print}" N
       "  " (list 'Atom= (fact-name associated-atom))
	    " {" (ncons (fact-name associated-atom)) " print}" N)
  (mapcar #'(lambda (link)
	      (msg "  " link " /" (link-node link) " cvx" N))
	  alist)
  (msg "] /new DefaultMenu send put" N)
  (msg name " /Clockwise false put" N)
  (cond ((eq name '*KB*) (msg "/ClientMenu *KB* def" N)))
  (msg EndPS N))

(defun link-name (link)
  (car link))

(defun link-node (link)
   (cadr link))

(defun follow (sexp node)
    (cond
      ((null node) nil)
      ((null sexp) (find-link '*NIL* node))
      ((atom sexp) (find-link sexp node))
      (t (follow (cdr sexp)
	         (follow (car sexp)
			 (find-link '*CONS* node))))))

(defun match-rest (rest node)
  (cond ((null rest) t)
	(t (match-fetch (car rest) (cdr rest) node))))

(defun match-var-rest (alist rest node)
  (cond ((null alist) nil)
	((member (link-name (car alist)) '(*CONS* *NIL*))
	 (match-var-rest (cdr alist) rest node))
	((match-rest rest (link-node (car alist))))
	(t (match-var-rest (cdr alist) rest node))))

(defun match-fetch (sexp rest node)
    (cond
      ((null node) nil)
      ((null sexp)
       (match-rest rest (find-link '*NIL* node)))
      ((atom sexp)
       (match-rest rest (find-link sexp node)))
      ((var-p sexp)
       (match-var-rest (node-alist node) rest node))
      (t (match-fetch (car sexp) (cons (cdr sexp) rest)
		      (find-link '*CONS* node)))))

(defun find-link (name node)
  (let ((link (assoc name (node-alist node))))
    (cond ((null link) nil)
	  (t (link-node link)))))

(defun fetch-nodes (sexp rest node bindings)
    (cond
      ((null node) nil)
      ((null (node-alist node))
       (ncons (cons (associated-fact node)
		    bindings)))
      ((null sexp)
       (fetch-nodes (car rest) (cdr rest)
		    (find-link '*NIL* node) bindings))
      ((atom sexp)
       (fetch-nodes (car rest) (cdr rest)
		    (find-link sexp node) bindings))
      ((var-p sexp)
       (cond ((assoc (var-name sexp) bindings)
	      (fetch-nodes (car rest) (cdr rest)
			   (find-link
			     (binding-val
			       (assoc (var-name sexp) bindings))
			     node)
			   bindings))
	     (t
	      (mapcan #'(lambda (link)
			  (and (not (member (link-name link)
					    '(*CONS* *NIL*)))
			       (fetch-nodes (car rest) (cdr rest)
					    (link-node link)
					    (cons
					      (list (var-name sexp)
						    (link-name link))
					      bindings))))
		      (node-alist node)))))
      (t (fetch-nodes (car sexp) (cons (cdr sexp) rest)
		      (find-link '*CONS* node) bindings))))

(defun fetch (sexp node)
  (fetch-nodes sexp nil node nil))

(defun fetch-fact-name (sexp node)
  (let ((facts (fetch sexp node)))
    (and facts (fact-name (car facts)))))


(defun change (sexp node atom)
  (let ((leaf-node (follow sexp node)))
    (cond (leaf-node (node-set-associated-atom leaf-node atom))
          (t nil))))

(defun associated-fact (node)
  (fact-name (node-associated-atom node)))

(defun fact-name (list)
       (car list))

(defun fact-bindings (list)
       (cdr list))

(defun get-status (fact)
       (get fact 'STATUS))

(defun set-status (fact status)
       (putprop fact status 'STATUS))

(defun ASSERT (fact &optional name)
       (cond ((null name)
	      (setq name (newsym 'FACT)))) ; make up a name if one not supplied
       (index fact '*KB* name)
       (set-status name 'KNOWN)
       name)

(defun STATUS (fact)
       (cond ((not (atom fact))
	      (setq fact (fetch-fact-name fact '*KB*))))
       (get-status fact))

(defun rule-set-fact (rule fact)
       (putprop rule fact 'FACT))

(defun rule-fact (rule)
       (get rule 'FACT))

(defun rule-set-prove (rule prove)
       (putprop rule prove 'PROVE))

(defun rule-prove (rule)
       (get rule 'PROVE))

(setq *RULE-LIST* nil)

(defun ASSERT-RULE (fact prove)
       (let ((rule-name (newsym 'RULE)))
	 (rule-set-fact rule-name (unabbreviate-rule fact))
	 (rule-set-prove rule-name (unabbreviate-rule prove))
	 (setq *RULE-LIST* (cons rule-name *RULE-LIST*))
	 rule-name))

(defun REBUILD (file)
  (node-reset-sexp '*KB*)
  (node-reset-alist '*KB*)
  (node-reset-associated-atom '*KB*)
  (setq *RULE-LIST* nil)
  (load file))

(defun SHOW-FACTS (&optional facts)
  (find-facts facts '*KB*))

(defun find-facts (facts node)
  (cond
    ((null (node-alist node))
     (cond
       ((or (not facts) (member (associated-fact node) facts))
	(msg "NAME: " (associated-fact node)
	", " (node-sexp node)
	", STATUS: " (get-status (associated-fact node))
	"." N)
       nil)))
    (t (mapcar #'(lambda (link)
		   (find-facts facts (link-node link)))
	       (node-alist node))
       nil)))

(defun SHOW-RULES (&optional rules)
  (mapcar #'(lambda (rule)
	      (msg  rule ": "
		    "I can prove " (abbreviate-rule (rule-fact rule))
		    " if I can prove " (abbreviate-rule (rule-prove rule))
		    "." N))
    (or rules *RULE-LIST*))
  nil)

(defun var-name (element)
  (cadr element))

(defun var-p (element)
  (and (listp element) (eq (car element) '*VAR*)))

(defun abbreviate-rule (rule)
  (cond ((null rule)
	 nil)
	((atom rule)
	 rule)
	((var-p rule)
	 (implode (cons '\? (explode (var-name rule)))))
	(t
	 (cons (abbreviate-rule (car rule))
	       (abbreviate-rule (cdr rule))))))

(defun unabbreviate-rule (rule)
  (cond ((null rule)
	 nil)
	((atom rule)
	 (let ((chars (explode rule)))
	    (cond ((eq (car chars) '?)
		   (list '*VAR* (implode (cdr chars))))
		  (t rule))))
	(t
	 (cons (unabbreviate-rule (car rule))
	       (unabbreviate-rule (cdr rule))))))

(defun match-rule (fact-pat fact bindings)
;  (msg "fact-pat: " fact-pat " fact: " fact " bindings: " bindings N)
  (cond
    ((and (null fact-pat) (null fact)) (or bindings t))
    ((or (null fact-pat) (null fact)) nil)
    ((atom fact-pat)
     (cond ((eq fact-pat fact) (or bindings t))
	   (t nil)))
    ((var-p fact-pat)
     (add-binding (var-name (cond ((var-p fact) fact) (t fact-pat)))
		  fact bindings))
    (t (let ((new-bindings (match-rule (car fact-pat) (car fact) bindings)))
	 (cond (new-bindings
		 (match-rule
		   (cdr fact-pat) (cdr fact)
		   (cond ((eq new-bindings t) nil)
			 (t new-bindings))))
	       (t nil))))))

(defun binding-val (binding)
  (cadr binding))

(defun add-binding (name val bindings)
  (let ((binding (assoc name bindings)))
    (cond (binding
	   (and (eq (binding-val binding) val)
                bindings))
	  (t (cons (list name val) bindings)))))

(defun match-facts (fact prove rule)
    (let ((result (PROVE prove)))
      (msg "fact: " fact " prove: " prove " result: " result N)
      (cond
	(result
	 (append
	   (cond ((eq result 'KNOWN) nil)
		 (t result))
	   (mapcan #'(lambda (known-fact)
		       (let ((conclusion (instantiate
					   fact
					   (fact-bindings known-fact))))
		         (cond
			   ((follow conclusion '*KB*) nil) ; already known?
			   (t (let ((new-fact (ASSERT conclusion)))
			        (msg "Conclusion: " conclusion
				     " new-fact: " new-fact N)
				(set-status new-fact
					    (list 'INFERRED rule))
				(ncons new-fact))))))
		   (fetch prove '*KB*))))
	(t nil))))

(defun prove1 (conjecture new-facts rules)
  (and
    rules
    (let ((fact-bindings (match-rule conjecture
				     (rule-fact (car rules)) nil)))
      (cond
	((not fact-bindings)
	 (prove1 conjecture new-facts (cdr rules)))
	(t
	  (msg "Rule-fact " (rule-fact (car rules)) N
	       "Rule-prove " (rule-prove (car rules)) N
	       "Fact-bindings " fact-bindings N)
	  (let* ((guess (instantiate (rule-prove (car rules))
				     fact-bindings))
		 (foo (msg "Guessing " guess N))
		 (result (PROVE guess)))
	     (msg "Result of " guess " is " result N)
	     (cond
	       ((not result)
		(prove1 conjecture new-facts (cdr rules)))
	       (t
		 (let ((matching-facts
			 (fetch (instantiate (rule-fact (car rules))
					     fact-bindings)
				'*KB*)))
		   (prove1 conjecture
			   (append
			     (mapcan '#(lambda (fact-bindings)
				         (let ((new (ASSERT
						      (instantiate
							(rule-fact (car rules))
							(cdr fact-bindings)))))
					   (set-status new
						       (list 'INFERRED rule))
					   new))
				     matching-facts)
			     new-facts)
			   (cdr rules)))))))))))

(defun prove2 (conjecture new-facts rules facts)
  (cond
    ((null facts) (prove1 conjecture new-facts (cdr rules)))
    (t (let ((new-conjecture (instantiate (rule-prove (car rules))
					  (fact-bindings (car facts)))))
		   (foo (msg "Conjecture: " conjecture N))
		   (conclusion (prove conjecture new-facts)))
		(msg "Conclusion: " conclusion N)
		(cond
		  ((fetch conjecture '*KB*)
		   (prove1 fact (cons
				  (ASSERT (instantiate (rule-fact rule)
						       fact-bindings))
				  conclusion)
			   (cdr rules)))
		  (t (prove1 fact new-facts (cdr rules)))
		  
		   ))
				(ASSERT (rule-fact rule))
				new-facts))
  (comment
	      (set-status (car new-facts) (list 'INFERRED rule)))

(defun prove (fact new-facts)
  (cond
    ((match-fetch fact () '*KB*)
     'KNOWN)
    (t (prove1 fact () *RULE-LIST*))))

(defun PROVE (fact)
;  (prove fact nil)
  (setq fact (unabbreviate-rule fact))
  (cond 
    ((follow fact '*KB*) 'KNOWN)
    ((mapcan #'(lambda (rule)
		 (let ((bindings (match-rule (rule-fact rule) fact nil)))
		   (cond (bindings
			  (match-facts
			    (instantiate (rule-fact rule)
					 bindings)
			    (instantiate (rule-prove rule)
					 bindings)
			    rule))
			 (t nil))))
	     *RULE-LIST*))
    ((fetch fact '*KB*) 'KNOWN)
    (t nil)))

(defun instantiate (pat bindings)
  (cond
    ((null pat) nil)
    ((atom pat) pat)
    ((var-p pat)
     (cond ((binding-val (assoc (var-name pat) bindings)))
	   (t pat)))
    (t (cons (instantiate (car pat) bindings)
	     (instantiate (cdr pat) bindings)))))

(REBUILD 'in)
